use embedded_graphics::{
    geometry::{Point, Size},
    mono_font::{ascii::FONT_10X20, MonoTextStyleBuilder},
    pixelcolor::{Rgb565, RgbColor},
    primitives::{Primitive, PrimitiveStyleBuilder, Rectangle},
    text::{Baseline, Text},
    Drawable,
};

use crate::keyboard::KeyCode;

use super::Element;

const CHAR_WIDTH: u32 = 10;
const CHAR_HEIGHT: u32 = 20;
const BORDER_WIDTH: u32 = 4;
const BACKGROUND_COLOR: Rgb565 = Rgb565::new(31, 41, 0); // #FFA500
const BORDER_COLOR: Rgb565 = Rgb565::new(31, 26, 0); // 0xFF6600

pub struct Button {
    x: i32,
    y: i32,

    text: &'static str,
    pub selected: bool,
}

impl Button {
    pub fn new(x: i32, y: i32, text: &'static str) -> Self {
        Self {
            x,
            y,
            text,
            selected: false,
        }
    }
}

impl Element for Button {
    type KeyPushReturn = bool;

    fn render<
        E,
        DT: embedded_graphics::prelude::DrawTarget<
            Color = embedded_graphics::pixelcolor::Rgb565,
            Error = E,
        >,
    >(
        &mut self,
        dt: &mut DT,
    ) -> Result<(), E> {
        let width = CHAR_WIDTH * u32::try_from(self.text.len()).unwrap() + BORDER_WIDTH * 2;
        let height = CHAR_HEIGHT + BORDER_WIDTH * 2;

        // Draw border/background
        let unselected_rect_style = PrimitiveStyleBuilder::new()
            .stroke_color(BORDER_COLOR)
            .stroke_width(BORDER_WIDTH)
            .fill_color(BACKGROUND_COLOR)
            .build();

        let selected_rect_style = PrimitiveStyleBuilder::new()
            .stroke_color(BORDER_COLOR)
            .stroke_width(BORDER_WIDTH)
            .fill_color(Rgb565::WHITE)
            .build();

        let rect_style = if self.selected {
            selected_rect_style
        } else {
            unselected_rect_style
        };

        Rectangle::new(Point::new(self.x, self.y), Size::new(width, height))
            .into_styled(rect_style)
            .draw(dt)?;

        // Draw text
        let text_x = self.x + i32::try_from(BORDER_WIDTH).unwrap();
        let text_y = self.y + i32::try_from(BORDER_WIDTH).unwrap();

        let text_style = MonoTextStyleBuilder::new()
            .font(&FONT_10X20)
            .text_color(Rgb565::BLACK)
            .build();

        Text::with_baseline(
            self.text,
            Point::new(text_x, text_y),
            text_style,
            Baseline::Top,
        )
        .draw(dt)?;

        Ok(())
    }

    fn key_push(&mut self, k: KeyCode) -> bool {
        k == KeyCode::Touchpad
    }
}
