mod contact;
mod message;

use contact::ContactGroup;
pub use contact::{
    Contact, ContactIter, MAX_CONTACTS, MAX_CONTACT_CALLSIGN_LENGTH, MAX_CONTACT_NAME_LENGTH,
};
pub use message::{Message, MessageDirection, MessageGroup, MessageIter};
use sectorize::SectorizeIter;

use core::slice::from_raw_parts;
use rp2040_flash::flash;

use self::message::AddresseeIter;

const SECTOR_SIZE: usize = 4096;
const FLASH_START: usize = 0x10000000;
// First 6MB is for code
const DATA_START: usize = FLASH_START + 6144 * 1024;
// 2MB for data
const DATA_END: usize = DATA_START + 2048 * 1024;

const X25: crc::Crc<u16> = crc::Crc::<u16>::new(&crc::CRC_16_IBM_SDLC);

pub struct Storage {
    contacts: ContactGroup,
    messages: MessageGroup,
}

impl Storage {
    pub fn new() -> Self {
        Self {
            contacts: ContactGroup::new(),
            messages: MessageGroup::new(),
        }
    }

    // Punch through our contact methods

    pub fn new_contact(&self) -> Contact {
        self.contacts.new_contact()
    }

    pub fn save_contact(&mut self, contact: Contact) {
        let mut contacts = self.contacts;
        contacts.save_contact(self, contact);
        self.contacts = contacts;
    }

    pub fn delete_contact(&mut self, contact_id: usize) {
        let mut contacts = self.contacts;
        contacts.delete_contact(self, contact_id);
        self.contacts = contacts;
    }

    pub fn contacts(&self) -> ContactIter {
        self.contacts.contacts()
    }

    pub fn contacts_len(&self) -> usize {
        self.contacts.len()
    }

    // Punch through our message methods

    pub fn push_message(&mut self, message: Message) {
        let mut messages = self.messages;
        messages.push_message(self, message);
        self.messages = messages;
    }

    pub fn messages<'a, 'c>(&'a self, callsign: &'c str, ssid: u8) -> MessageIter<'a, 'a, 'c> {
        self.messages.iter(self, callsign, ssid)
    }

    pub fn message_addressees(&self) -> AddresseeIter {
        self.messages.iter_addressees(self)
    }

    fn write_data(&mut self, start_addr: usize, buf: &[u8]) {
        // rprintln!("Write {:x?} to {:x}", buf, start_addr);

        for s in SectorizeIter::new(SECTOR_SIZE, start_addr, buf.len()) {
            let mut existing = self.read_slice(s.sector_index * SECTOR_SIZE);

            /* rprintln!(
                    "Sector {}: {}..{} <- {}..{}",
                    s.sector_index,
                    s.sector_start,
                    s.sector_end,
                    s.input_start,
                    s.input_end
            ); */

            // only write if there's a difference
            if existing[s.sector_start..s.sector_end] != buf[s.input_start..s.input_end] {
                existing[s.sector_start..s.sector_end]
                    .copy_from_slice(&buf[s.input_start..s.input_end]);
                self.write_sector(s.sector_index * SECTOR_SIZE, &existing);
            }
        }
    }

    fn read_slice<const N: usize>(&self, addr: usize) -> [u8; N] {
        assert!(addr >= DATA_START);
        assert!(addr + N < DATA_END);

        let sector = unsafe { from_raw_parts(addr as *mut u8, N) };

        let mut buf = [0; N];

        buf.copy_from_slice(sector);

        buf
    }

    fn write_sector(&mut self, start_addr: usize, buf: &[u8; 4096]) {
        let addr: u32 = (start_addr - FLASH_START).try_into().unwrap();
        let sector_size = u32::try_from(SECTOR_SIZE).unwrap();
        assert!(addr % sector_size == 0);

        cortex_m::interrupt::free(|_| unsafe {
            flash::flash_range_erase_and_program(addr, buf, true);
        });
    }
}
